package java.rmi.server;

import sun.rmi.server.Util;
import java.util.WeakHashMap;
import sun.rmi.server.WeakClassHashMap;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.Map;
import java.io.InvalidObjectException;
import java.rmi.UnexpectedException;
import java.rmi.RemoteException;
import java.rmi.Remote;
import java.lang.reflect.Proxy;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationHandler;

public class RemoteObjectInvocationHandler extends RemoteObject implements InvocationHandler
{
    private static final long serialVersionUID = 2L;
    private static final boolean allowFinalizeInvocation;
    private static final RemoteObjectInvocationHandler.MethodToHash_Maps methodToHash_Maps;
    
    public RemoteObjectInvocationHandler(RemoteRef remoteRef) {
        super(remoteRef);
        if (remoteRef == null) {
            throw new NullPointerException();
        }
        return;
    }
    
    @Override
    public Object invoke(Object proxy, Method method, Object[] array) throws Throwable {
        if (!Proxy.isProxyClass(proxy.getClass())) {
            throw new IllegalArgumentException("not a proxy");
        }
        if (Proxy.getInvocationHandler(proxy) != this) {
            throw new IllegalArgumentException("handler mismatch");
        }
        if (method.getDeclaringClass() == Object.class) {
            return this.invokeObjectMethod(proxy, method, array);
        }
        if ("finalize".equals(method.getName()) && method.getParameterCount() == 0 && !RemoteObjectInvocationHandler.allowFinalizeInvocation) {
            return null;
        }
        return this.invokeRemoteMethod(proxy, method, array);
    }
    
    private Object invokeObjectMethod(Object o, Method obj, Object[] array) {
        String name = obj.getName();
        if (name.equals("hashCode")) {
            return Integer.valueOf(this.hashCode());
        }
        if (name.equals("equals")) {
            Object proxy = array[0];
            InvocationHandler invocationHandler;
            return Boolean.valueOf(o == proxy || (proxy != null && Proxy.isProxyClass(proxy.getClass()) && (invocationHandler = Proxy.getInvocationHandler(proxy)) instanceof RemoteObjectInvocationHandler && this.equals(invocationHandler)));
        }
        if (name.equals("toString")) {
            return this.proxyToString(o);
        }
        throw new IllegalArgumentException("unexpected Object method: " + obj);
    }
    
    private Object invokeRemoteMethod(Object o, Method method, Object[] array) throws Exception {
        try {
            if (!(o instanceof Remote)) {
                throw new IllegalArgumentException("proxy not Remote instance");
            }
            Class<Object> declaringClass = method.getDeclaringClass();
            if (!Remote.class.isAssignableFrom(declaringClass)) {
                throw new RemoteException("Method is not Remote: " + declaringClass + "::" + method);
            }
            return this.ref.invoke((Remote)o, method, array, getMethodHash(method));
        }
        catch (Exception ex) {
            if (!(ex instanceof RuntimeException)) {
                Class<Object> class1 = o.getClass();
                try {
                    method = class1.getMethod(method.getName(), method.getParameterTypes());
                }
                catch (NoSuchMethodException cause) {
                    throw (IllegalArgumentException)new IllegalArgumentException().initCause(cause);
                }
                Class<Object> class2 = ex.getClass();
                Class[] exceptionTypes = method.getExceptionTypes();
                int length = exceptionTypes.length;
                int i = 0;
                while (i < length) {
                    if (exceptionTypes[i].isAssignableFrom(class2)) {
                        throw ex;
                    }
                    ++i;
                }
                ex = new UnexpectedException("unexpected exception", ex);
            }
            throw ex;
        }
    }
    
    private String proxyToString(Object o) {
        Class[] interfaces = o.getClass().getInterfaces();
        if (interfaces.length == 0) {
            return "Proxy[" + this + "]";
        }
        String str = interfaces[0].getName();
        if (str.equals("java.rmi.Remote") && interfaces.length > 1) {
            str = interfaces[1].getName();
        }
        int lastIndex = str.lastIndexOf(46);
        if (lastIndex >= 0) {
            str = str.substring(lastIndex + 1);
        }
        return "Proxy[" + str + "," + this + "]";
    }
    
    private void readObjectNoData() throws InvalidObjectException {
        throw new InvalidObjectException("no data in stream; class: " + this.getClass().getName());
    }
    
    private static long getMethodHash(Method method) {
        return (long)((Map)RemoteObjectInvocationHandler.methodToHash_Maps.get(method.getDeclaringClass())).get(method);
    }
    
    static {
        String s = (String)AccessController.doPrivileged(new PrivilegedAction<String>() {
            @Override
            public String run() {
                return System.getProperty("sun.rmi.server.invocationhandler.allowFinalizeInvocation");
            }
        });
        if ("".equals(s)) {
            allowFinalizeInvocation = true;
        }
        else {
            allowFinalizeInvocation = Boolean.parseBoolean(s);
        }
        methodToHash_Maps = new RemoteObjectInvocationHandler.MethodToHash_Maps();
    }
    
    private static class MethodToHash_Maps extends WeakClassHashMap<Map<Method, Long>>
    {
        MethodToHash_Maps() {
        }
        
        @Override
        protected Map<Method, Long> computeValue(Class<?> clazz) {
            return new WeakHashMap<Method, Long>() {
                @Override
                public synchronized Long get(Object key) {
                    Long value = (Long)super.get(key);
                    if (value == null) {
                        Method method = (Method)key;
                        value = Long.valueOf(Util.computeMethodHash(method));
                        this.put(method, value);
                        return value;
                    }
                    return value;
                }
            };
        }
    }
}
